
#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

import logging
import os
import shutil

_log = logging.getLogger(__name__)


# Represents a coding rule violation
class RuleViolation(object):
    def __init__(self, rule, scope, details = None):
        self.rule       = rule
        self.scope      = scope
        self.details    = details

class FileRuleViolation(RuleViolation):
    def __init__(self, rule, scope, details = None,
                 line = None, function = None, class_ = None):
        RuleViolation.__init__(self, rule, scope, details)
        self.line           = line
        self.function       = function
        self.class_         = class_


# Represents a quality metric measurement
class MetricMeasurement(object):
    def __init__(self, metric, scope, value):
        self.metric = metric
        self.scope  = scope
        self.value  = value


class FileMetricMeasurement(MetricMeasurement):
    def __init__(self, metric, scope, value,
                 line = None, function = None, class_ = None):
        MetricMeasurement.__init__(self, metric, scope, value)
        self.line           = line
        self.function       = function
        self.class_         = class_


class UndefinedPropertyError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class AnalysisScopeError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

# Provides an interface for plugins to communicate with the framework.
class PluginInterface(object):
    def __init__(self, data):
        self.state          = None
        self._data          = data
        self._plugin        = None
        self._scope         = None
        self._exported      = set()

    def get_file(self, relative_path):
        return os.path.join(self._plugin.path, relative_path)

    def export_file(self, relative_path):
        # mark a file in the plugin's temporary directory as exportable
        target = os.path.join(self._plugin.tmp_path, relative_path)
        if os.path.isfile(target):
            self._exported.add(target)

    def report_file_violation(self, rule_id, msg, scope_id,
                              line = None, function = None, class_ = None):
        _log.debug("file_violation(%s, %s, %s)", rule_id, msg, scope_id)
        scope = self._get_scope(scope_id, self._data.files)
        r = self._get_property(rule_id, self._data.rules, scope)
        datum = FileRuleViolation(r, scope, msg, line, function, class_)
        scope._violations.append(datum)

    def report_package_violation(self, rule_id, msg, scope_id):
        _log.debug("package_violation(%s, %s, %s)", rule_id, msg, scope_id)
        scope = self._get_scope(scope_id, self._data.packages)
        r = self._get_property(rule_id, self._data.rules, scope)
        scope._violations.append(RuleViolation(r, scope, msg))

    def report_repository_violation(self, rule_id, msg, scope_id):
        _log.debug("repository_violation(%s, %s, %s)", rule_id, msg, scope_id)
        scope = self._get_scope(scope_id, self._data.repositories)
        r = self._get_property(rule_id, self._data.rules, scope)
        scope._violations.append(RuleViolation(r, scope, msg))

    # Shorthand for reporting a violation on the current scope
    def report_violation(self, rule_id, msg,
                         line = None, function = None, class_ = None):
        _log.debug("violation(%s, %s, %s)", rule_id, msg, self._scope.id)
        r = self._get_property(rule_id, self._data.rules, self._scope)
        if r.scope == "repository" or r.scope == "package":
            datum = RuleViolation(r, self._scope, msg)
        else:
            datum = FileRuleViolation(r, self._scope, msg,
                                      line, function, class_)
        self._scope._violations.append(datum)

    def report_file_metric(self, metric_id, value, scope_id,
                           line = None, function = None, class_ = None):
        _log.debug("file_metric(%s, %s, %s)", metric_id, value, scope_id)
        scope = self._get_scope(scope_id, self._data.files)
        m = self._get_property(metric_id, self._data.metrics, scope)
        datum = FileMetricMeasurement(m, scope, value, line, function, class_)
        scope._metrics.append(datum)
        self._check_metric_violation(datum)

    def report_package_metric(self, metric_id, value, scope_id):
        _log.debug("package_metric(%s, %s, %s)", metric_id, value, scope_id)
        scope = self._get_scope(scope_id, self._data.packages)
        m = self._get_property(metric_id, self._data.metrics, scope)
        datum = MetricMeasurement(m, scope, value)
        scope._metrics.append(datum)
        self._check_metric_violation(datum)

    def report_repository_metric(self, metric_id, value, scope_id):
        _log.debug("repository_metric(%s, %s, %s)", metric_id, value, scope_id)
        scope = self._get_scope(scope_id, self._data.repositories)
        m = self._get_property(metric_id, self._data.metrics, scope)
        datum = MetricMeasurement(m, scope, value)
        scope._metrics.append(datum)
        self._check_metric_violation(datum)

    # Shorthand for reporting a metric on the current scope
    def report_metric(self, metric_id, value,
                      line = None, function = None, class_ = None):
        _log.debug("metric(%s, %s, %s)", metric_id, value, self._scope.id)
        m = self._get_property(metric_id, self._data.metrics, self._scope)
        if m.scope == "repository" or m.scope == "package":
            datum = MetricMeasurement(m, self._scope, value)
        else:
            datum = FileMetricMeasurement(m, self._scope, value,
                                          line, function, class_)
        self._scope._metrics.append(datum)
        self._check_metric_violation(datum)

    def register_configuration(self, configuration):
        _log.debug("config(%s)", configuration)
        scope = self._get_scope(configuration.package, self._data.packages)
        scope._configs.append(configuration)

    def find_package(self, scope_id):
        return self._data.packages.get(scope_id, None)

    def _get_scope(self, scope_id, data):
        if not scope_id in data:
            raise AnalysisScopeError("Unknown scope id " + scope_id)
        scope = data[scope_id]
        if not self._scope is None and not self._scope.bound_to(scope):
            raise AnalysisScopeError("Unrelated scope " + scope_id)
        return scope

    def _get_property(self, property_id, data, scope):
        id = property_id
        if not property_id in data:
            id = self._plugin.name + ":" + property_id
            if not id in data:
                raise UndefinedPropertyError(property_id)
        datum = data[id]
        if not scope.accepts_scope(datum.scope):
            raise AnalysisScopeError("Found " + datum.scope
                                     + "; Expected " + scope.scope)
        return datum

    def _check_metric_violation(self, measurement):
        _log.debug("_check_metric_violation(%s)", measurement.metric.id)
        tmax = measurement.metric.maximum
        tmin = measurement.metric.minimum
        value = measurement.value
        if (not tmax is None and value > tmax) \
                or (not tmin is None and value < tmin):
            rule_id = "metric:" + measurement.metric.id
            if rule_id in self._data.rules:
                rule = self._data.rules[rule_id]
                msg = "Reported metric value: " + str(value)
                if rule.scope == "repository" or rule.scope == "package":
                    datum = RuleViolation(rule, measurement.scope, msg)
                else:
                    datum = FileRuleViolation(rule, measurement.scope, msg,
                                              measurement.line,
                                              measurement.function,
                                              measurement.class_)
                measurement.scope._violations.append(datum)



def run_analysis_and_processing(datadir, plugins, data, expodir):
    _log.info("Running plugins on collected data.")
    iface = PluginInterface(data)
    # Step 0: prepare directories
    plugout = os.path.join(datadir, ".plugout")
    if os.path.exists(plugout):
        shutil.rmtree(plugout)
    os.mkdir(plugout)
    for plugin in plugins:
        path = os.path.join(plugout, plugin.name)
        os.mkdir(path)
        plugin.tmp_path = path
    wd = os.getcwd()
    try:
        _analysis(iface, plugins, data)
        _processing(iface, plugins, data)
        _exports(iface._exported, expodir)
    finally:
        os.chdir(wd)
        shutil.rmtree(plugout)
    


def _analysis(iface, plugins, data):
    for plugin in plugins:
        _log.debug("Running analyses for " + plugin.name)
        os.chdir(plugin.tmp_path)
        # Step 1: run initialisation
        plugin.analysis.pre_analysis()
        # Step 2: run analysis; file > package > repository
        iface._plugin = plugin
        iface.state = plugin.analysis.state
        try:
            for scope in data.files.itervalues():
                if (data.launch_files and scope.language == "launch"
                                      and not scope in data.launch_files):
                    continue
                iface._scope = scope
                plugin.analysis.analyse_file(iface, scope)
            for scope in data._topological_packages:
                iface._scope = scope
                plugin.analysis.analyse_package(iface, scope)
            for scope in data.repositories.itervalues():
                iface._scope = scope
                plugin.analysis.analyse_repository(iface, scope)
        except (UndefinedPropertyError, AnalysisScopeError) as e:
            _log.error("%s", e.value)
        # Step 3: run finalisation
        iface._scope = None
        plugin.analysis.post_analysis(iface)


def _processing(iface, plugins, data):
    # Step 1: run initialisation
    for plugin in plugins:
        os.chdir(plugin.tmp_path)
        plugin.process.pre_process()
    # Step 2: run processing; file > package > repository
    try:
        for scope in data.files.itervalues():
            iface._scope = scope
            v = list(scope._violations)
            m = list(scope._metrics)
            for plugin in plugins:
                iface._plugin = plugin
                iface.state = plugin.process.state
                os.chdir(plugin.tmp_path)
                plugin.process.process_file(iface, scope, v, m)
        for scope in data._topological_packages:
            iface._scope = scope
            v = list(scope._violations)
            m = list(scope._metrics)
            for plugin in plugins:
                iface._plugin = plugin
                iface.state = plugin.process.state
                os.chdir(plugin.tmp_path)
                plugin.process.process_package(iface, scope, v, m)
        for scope in data.repositories.itervalues():
            iface._scope = scope
            v = list(scope._violations)
            m = list(scope._metrics)
            for plugin in plugins:
                iface._plugin = plugin
                iface.state = plugin.process.state
                os.chdir(plugin.tmp_path)
                plugin.process.process_repository(iface, scope, v, m)
    except (UndefinedPropertyError, AnalysisScopeError) as e:
        _log.error("%s", e.value)
    # Step 3: run finalisation
    iface._scope = None
    for plugin in plugins:
        os.chdir(plugin.tmp_path)
        plugin.process.post_process(iface)


def _exports(sources, expodir):
    counter = 1
    for f in sources:
        target = os.path.join(expodir, f[f.rfind(os.sep)+1:])
        if os.path.isfile(target):
            i = f.rfind(os.sep)
            j = f.rfind(".")
            if i >= j:
                ext = ".data"
            else:
                ext = f[j:]
            name = "d{:04d}{}".format(counter, ext)
            target = os.path.join(expodir, name)
            counter += 1
        shutil.move(f, target)




#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

import cPickle
import logging
from operator import attrgetter
import os
import subprocess
import yaml
import xml.etree.ElementTree as ET

import urllib
try:
    from urllib.request import urlopen
    from urllib.error import HTTPError
except ImportError:
    from urllib2 import urlopen
    from urllib2 import HTTPError

from rospkg import RosPack, ResourceNotFound


SCOPE_TYPES = ("function", "class", "file", "package", "repository")

_log = logging.getLogger(__name__)


###############################################################################
# Analysis Properties
###############################################################################

# Represents a coding rule
class Rule(object):
    def __init__(self, rule_id, name, scope, desc, tags):
        self.id             = rule_id
        self.name           = name
        assert scope in SCOPE_TYPES
        self.scope          = scope
        self.description    = desc
        self.tags           = tags


# Represents a quality metric
class Metric(object):
    def __init__(self, metric_id, name, scope, desc, minv = None, maxv = None):
        self.id             = metric_id
        self.name           = name
        assert scope in SCOPE_TYPES
        self.scope          = scope
        self.description    = desc
        self.minimum        = minv
        self.maximum        = maxv


###############################################################################
# Source Code Structures
###############################################################################

# Base class for general utility
class AnalysisScope(object):
    def __init__(self, id, name, scope):
        self.id     = id
        self.name   = name
        assert scope in SCOPE_TYPES
        self.scope  = scope

    def lte_scope(self, scope):
        return SCOPE_TYPES.index(self.scope) <= SCOPE_TYPES.index(scope)

    def gte_scope(self, scope):
        return SCOPE_TYPES.index(self.scope) >= SCOPE_TYPES.index(scope)

    def lt_scope(self, scope):
        return SCOPE_TYPES.index(self.scope) < SCOPE_TYPES.index(scope)

    def gt_scope(self, scope):
        return SCOPE_TYPES.index(self.scope) > SCOPE_TYPES.index(scope)

    def bound_to(self, scope):
        return self == scope

    def accepts_scope(self, scope):
        return self.scope == scope

    def __str__(self):
        return self.id


# Represents a person (author/maintainer)
class Person(object):
    def __init__(self, name, email = None):
        self.name = name if name else "?"
        self.email = email

    def __eq__(self, other):
        if not type(self) is type(other):
            return False
        if self.email and other.email:
            return self.email == other.email
        return self.name == other.name

    def __hash__(self):
        return hash(self.email) if self.email else hash(self.name)


# Represents a source code file
class SourceFile(AnalysisScope):
    _id_gen         = 1
    _excluded_dirs  = [".git", "doc", "bin", "cmake"]
    _cpp_sources    = (".cpp", ".cc", ".h", ".hpp", ".c", ".cpp.in", ".h.in",
                       ".hpp.in", ".c.in", ".cc.in")
    _py_sources     = ".py"
    _manifests      = "package.xml"
    _launch_sources = ".launch"

    def __init__(self, name, path, pkg, lang):
        AnalysisScope.__init__(self, pkg.id + ":" + str(SourceFile._id_gen),
                                name, "file")
        SourceFile._id_gen += 1
        self.path       = path # relative to package root
        self.package    = pkg
        self.language   = lang
        full_path       = self.get_path()
        self.size       = os.path.getsize(full_path)
        self.lines      = SourceFile._count_physical_lines(full_path)
        self._violations = []
        self._metrics   = []

    @classmethod
    def populate_package(cls, pkg, idx = None):
        _log.debug("SourceFile.populate_package(%s)", pkg)
        if pkg.path is None:
            _log.debug("Package %s has no path", pkg)
            return
        _log.info("Indexing source files for package %s", pkg)
        prefix = len(pkg.path) + len(os.path.sep)
        for root, subdirs, files in os.walk(pkg.path, topdown = True):
            subdirs[:] = [d for d in subdirs if d not in cls._excluded_dirs]
            path = root[prefix:]
            for f in files:
                source = None
                if f.endswith(cls._cpp_sources):
                    source = "cpp"
                elif f.endswith(cls._py_sources):
                    source = "python"
                elif f == cls._manifests:
                    source = "package"
                elif f.endswith(cls._launch_sources):
                    source = "launch"
                if source:
                    _log.debug("Found %s file %s at %s", source, f, path)
                    source = cls(f, path, pkg, source)
                    pkg.source_files.append(source)
                    pkg.size += source.size
                    pkg.lines += source.lines
                    if not idx is None:
                        idx[source.id] = source

    def get_path(self):
        return os.path.join(self.package.path, self.path, self.name)

    @staticmethod
    def _count_physical_lines(path):
        _log.debug("SourceFile._count_physical_lines(%s)", path)
        count = 0
        with open(path, "r") as handle:
            for count, _ in enumerate(handle, start = 1):
                pass
        return count

    def bound_to(self, other):
        if other.scope == "package":
            return self.package == other
        if other.scope == "repository":
            return self.package in other.packages
        return self == other

    def accepts_scope(self, scope):
        return self.gte_scope(scope)


# Represents a ROS package
class Package(AnalysisScope):
    def __init__(self, name, repo = None):
        AnalysisScope.__init__(self, name, name, "package")
    # public:
        self.repository         = repo
        self.authors            = set()
        self.maintainers        = set()
        self.isMetapackage      = False
        self.description        = ""
        self.licenses           = set()
        self.dependencies       = set()
        self.website            = None
        self.vcs_url            = None
        self.bug_url            = None
        self.path               = None
        self.source_files       = []
        self.nodelets           = []
        self.size               = 0 # sum of file sizes
        self.lines              = 0 # sum of physical file lines
    # private:
        self._violations        = []
        self._metrics           = []
        self._configs           = []
        self._tier              = 0 # for topological sort

    @classmethod
    def from_manifest(cls, pkg_file, repo = None):
        _log.debug("Package.from_manifest(%s, %s)", pkg_file, repo)
        with open(pkg_file, "r") as handle:
            root = ET.parse(handle).getroot()
        name = root.find("name").text
        package = cls(name, repo)
        package.path = os.path.dirname(pkg_file)
        _log.info("Found package %s at %s", package, package.path)
        for el in root.findall("maintainer"):
            name = el.text
            email = el.get("email")
            if not name and not email:
                continue # No name, no email, move on
            package.maintainers.add(Person(name, email))
        for el in root.findall("author"):
            name = el.text
            email = el.get("email")
            if not name and not email:
                continue # No name, no email, move on
            package.authors.add(Person(name, email))
        el = root.find("export")
        if not el is None:
            if not el.find("metapackage") is None:
                package.isMetapackage = True
            if not el.find("nodelet") is None:
                nodelets = el.find("nodelet").get("plugin")
                nodelets = nodelets.replace("${prefix}", package.path)
                package._read_nodelets(nodelets)
        package.description = root.find("description").text.strip()
        for el in root.findall("license"):
            package.licenses.add(el.text)
        for el in root.findall("url"):
            value = el.get("type")
            if value is None or value == "website":
                package.website = el.text
            elif value == "repository":
                package.vcs_url = el.text
            elif value == "bugtracker":
                package.bug_url = el.text
        for el in root.findall("build_depend"):
            package.dependencies.add(el.text)
        if root.get("format") == "2":
            for el in root.findall("depend"):
                package.dependencies.add(el.text)
            for el in root.findall("build_export_depend"):
                package.dependencies.add(el.text)
            for el in root.findall("exec_depend"):
                package.dependencies.add(el.text)
        else:
            for el in root.findall("run_depend"):
                package.dependencies.add(el.text)
        return package

    @classmethod
    def locate_offline(cls, name, repo_path = None):
        _log.debug("Package.locate_offline(%s, %s)", name, repo_path)
        # Step 1: use repository download directory
        if repo_path:
            rp = RosPack.get_instance([repo_path])
            try:
                path = rp.get_path(name)
                path = os.path.join(path, "package.xml")
                if os.path.isfile(path):
                    _log.debug("Found %s in local repositories.", name)
                    return cls.from_manifest(path)
            except ResourceNotFound as e:
                _log.debug("%s was not found in local repositories.", name)
        # Step 2: use known directories
        rp = RosPack.get_instance()
        try:
            path = rp.get_path(name)
            path = os.path.join(path, "package.xml")
            if os.path.isfile(path):
                _log.debug("Found %s in default paths.", name)
                return cls.from_manifest(path)
        except ResourceNotFound as e:
            _log.debug("%s was not found in default paths.", name)
        return None

    # Note: this method messes with private variables of the RosPack
    # class. This is needed because, at some point, we download new
    # repositories and the package cache becomes outdated.
    # RosPack provides no public method to refresh the cache, hence
    # changing private variables directly.
    @staticmethod
    def refresh_package_cache(repo_path = None):
        if repo_path:
            rp = RosPack.get_instance([repo_path])
        else:
            rp = RosPack.get_instance()
        rp._location_cache = None

    def bound_to(self, other):
        if other.scope == "file":
            return other.package == self
        if other.scope == "repository":
            return self.repository == other
        return self == other or other.id in self.dependencies

    def __repr__(self):
        return "Package({})".format(self.id)

    def _read_nodelets(self, nodelet_plugins):
        _log.debug("Package._read_nodelets(%s)", nodelet_plugins)
        with open(nodelet_plugins, "r") as handle:
            root = ET.parse(handle).getroot()
        _log.info("Found nodelets at %s", nodelet_plugins)
        if root.tag == "library":
            libs = (root,)
        else:
            libs = root.findall("library")
        for el in libs:
            libname = el.get("path").rsplit(os.sep)[-1]
            for cl in el.findall("class"):
                self.nodelets.append((libname, cl.get("name")))



class RepositoryCloneError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class Repository(AnalysisScope):
    def __init__(self, name):
        AnalysisScope.__init__(self, name, name, "repository")
        self.vcs            = None 
        self.url            = None 
        self.version        = None
        self.status         = None
        self.path           = None
        self.packages       = []
        self.declared_packages = []
        self.commits        = 1
        self.contributors   = 1
        self._violations    = []
        self._metrics       = []

    def __eq__(self, other):
        if not type(self) is type(other):
            return False
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def download(self, repo_path):
        _log.debug("Repository.download(%s)", repo_path)
        if not self.url:
            _log.debug("%s has no URL to download from.", self.id)
            return
        path = os.path.join(repo_path, self.id)
        clone = False
        if not os.path.exists(path):
            os.makedirs(path)
            clone = True
        os.chdir(path)
        if self.vcs == "git":
            self._download_git(path, clone)
        elif self.vcs == "hg":
            self._download_hg(path, clone)
        elif self.vcs == "svn":
            self._download_svn(path, clone)

    def _download_git(self, path, clone = False):
        _log.debug("Repository._download_git(%s)", path)
        try:
            if clone:
                subprocess.check_call(["git", "init"])
                subprocess.check_call(["git", "remote", "add", "-t",
                        self.version, "-f", "origin", self.url])
                subprocess.check_call(["git", "checkout", self.version])
            else:
                subprocess.check_call(["git", "pull"])
            self.path = path
            self.commits = int(subprocess.check_output(["git", \
                    "rev-list", "HEAD", "--count"]).rstrip())
        except subprocess.CalledProcessError as e:
            raise RepositoryCloneError("git error")

    def _download_hg(self, path, clone = False):
        _log.debug("Repository._download_hg(%s)", path)
        try:
            if clone:
                subprocess.check_call(["hg", "clone", self.url, \
                        "-r", self.version])
            else:
                subprocess.check_call(["hg", "pull"])
            self.path = path
            self.commits = int(subprocess.check_output("hg", \
                    "id", "--num", "--rev", "tip").rstrip())
        except subprocess.CalledProcessError as e:
            raise RepositoryCloneError("hg error")

    def _download_svn(self, path, clone = False):
        _log.debug("Repository._download_svn(%s)", path)
        try:
            if clone:
                subprocess.check_call(["git", "svn", "clone", "-T", \
                        self.version if self.version == "trunk" \
                        else "branches/" + self.version, \
                        self.url])
            else:
                subprocess.check_call(["git", "svn", "fetch"])
            self.path = path
            self.commits = int(subprocess.check_output(["git", \
                    "rev-list", "HEAD", "--count"]).rstrip())
        except subprocess.CalledProcessError as e:
            raise RepositoryCloneError("git-svn error")

    @classmethod
    def from_distribution_data(cls, name, data):
        _log.debug("Repository.from_distribution_data(%s)", name)
        if not "source" in data:
            _log.debug("There is no source in provided data.")
            return None
        repo = cls(name)
        repo.status = data.get("status")
        src = data["source"]
        repo.vcs = src["type"]
        repo.url = src["url"]
        repo.version = src["version"]
        if "release" in data:
            repo.declared_packages = data["release"].get("packages", [])
        return repo

    @classmethod
    def from_user_data(cls, name, data):
        _log.debug("Repository.from_user_data(%s)", name)
        repo = cls(name)
        repo.status = "private"
        repo.vcs = data["type"]
        repo.url = data["url"]
        repo.version = data["version"]
        repo.declared_packages = data["packages"]
        return repo

    # User-defined repositories have priority.
    # Tries to build and save as few repositories as possible.
    @classmethod
    def load_repositories(cls, user_repos = None, pkg_list = None):
        _log.debug("Repository.load_repositories(%s)", pkg_list)
        repos = {}
        pkg_list = list(pkg_list) if pkg_list else None
        if user_repos:
            _log.info("Looking up user provided repositories.")
            if pkg_list is None:
                for id, info in user_repos.iteritems():
                    repos[id] = cls.from_user_data(id, info)
            else:
                for id, info in user_repos.iteritems():
                    if not pkg_list:
                        return repos
                    repo = cls.from_user_data(id, info)
                    for pkg in repo.declared_packages:
                        if pkg in pkg_list:
                            repos[id] = repo
                            pkg_list.remove(pkg)
        if not pkg_list:
            return repos
        url = "https://raw.githubusercontent.com/ros/rosdistro/master/" \
              + os.environ["ROS_DISTRO"] + "/distribution.yaml"
        _log.info("Looking up repositories from official distribution.")
        data = yaml.load(urlopen(url).read())["repositories"]
        if pkg_list is None:
            for id, info in data.iteritems():
                repo = cls.from_distribution_data(id, info)
                if repo:
                    repos[id] = repo
        else:
            for id, info in data.iteritems():
                if not pkg_list:
                    return repos
                repo = cls.from_distribution_data(id, info)
                if repo:
                    for pkg in repo.declared_packages:
                        if pkg in pkg_list:
                            repos[id] = repo
                            pkg_list.remove(pkg)
        return repos

    def bound_to(self, other):
        if other.scope == "package":
            return other.repository == self
        if other.scope == "file":
            return other.package in self.packages
        return self == other


################################################################################
# Data Manager
################################################################################

# Object to store and manage all data
class DataManager(object):
    def __init__(self):
        self.launch_files   = []
        self.repositories   = {}
        self.packages       = {}
        self.files          = {}
        self.rules          = {}
        self.metrics        = {}
        self.common_rules   = set()
        self.common_metrics = set()

        self._topological_packages = []

    def _add_metric(self, id, metric, common = False):
        minv = metric.get("min")
        maxv = metric.get("max")
        m = Metric(id, metric["name"], metric["scope"], metric["description"],
                   minv = float(minv) if not minv is None else None,
                   maxv = float(maxv) if not maxv is None else None)
        self.metrics[id] = m
        if common:
            self.common_metrics.add(id)
        id = "metric:" + id
        if (not m.minimum is None or not m.maximum is None) and \
                not id in self.rules:
            self.rules[id] = Rule(id, metric["name"] + " Threshold",
                                  metric["scope"],
                                  metric["name"] + " threshold is [" \
                                  + str(metric.get("min", "n/a")) + "," \
                                  + str(metric.get("max", "n/a")) + "]", \
                                  ["metrics"])
            if common:
                self.common_rules.add(id)

    # Used at startup to load the common rules and metrics
    def load_definitions(self, data_file):
        _log.debug("DataManager.load_definitions(%s)", data_file)
        with open(data_file, "r") as handle:
            data = yaml.load(handle)
        for id, rule in data.get("rules", {}).iteritems():
            self.rules[id] = Rule(id, rule["name"], rule["scope"],
                                  rule["description"], rule["tags"])
            self.common_rules.add(id)
        for id, metric in data.get("metrics", {}).iteritems():
            self._add_metric(id, metric, common = True)

    # Used to register the custom rules and metrics that each plugin defines
    def extend_definitions(self, plugin_id, rules, metrics):
        _log.debug("DataManager.extend_definitions(%s)", plugin_id)
        for id, rule in rules.iteritems():
            if id in self.common_rules:
                _log.warning("Plugin %s cannot override %s", plugin_id, id)
                continue # cannot override common rules
            id = plugin_id + ":" + id
            self.rules[id] = Rule(id, rule["name"], rule["scope"],
                                  rule["description"], rule["tags"])
        for id, metric in metrics.iteritems():
            if id in self.common_metrics:
                _log.warning("Plugin %s cannot override %s", plugin_id, id)
                continue # cannot override common metrics
            self._add_metric(plugin_id + ":" + id, metric)

    # Used at startup to build the list of packages, repositories
    # and source files that are to be analysed
    def index_source(self, index_file, repo_path = None, index_repos = False):
        _log.debug("DataManager.index_source(%s, %s)", index_file, repo_path)
        with open(index_file, "r") as handle:
            data = yaml.load(handle)
    # Step 1: find packages locally
        _log.info("Looking for packages locally.")
        missing = []
        pkg_list = self._read_launch_listing(data.get("launch"))
        pkg_list.extend(data.get("packages", []))
        pkg_list = set(pkg_list)
        for id in pkg_list:
            pkg = Package.locate_offline(id, repo_path)
            if pkg is None:
                missing.append(id)
            else:
                SourceFile.populate_package(pkg, self.files)
                self.packages[id] = pkg
    # Step 2: load repositories only if explicitly told to
        _log.debug("Missing packages: %s", missing)
        if index_repos:
            _log.info("Indexing repositories.")
            self.repositories = Repository.load_repositories(
                    data.get("repositories", {}), pkg_list)
            repos = set()
            for _, repo in self.repositories.iteritems():
                for id in repo.declared_packages:
                    if id in missing:
                        _log.debug("%s contains missing %s", _, id)
                        repos.add(repo)
    # Step 3: clone necessary repositories
            wd = os.getcwd()
            refresh = False
            for repo in repos:
                try:
                    repo.download(repo_path)
                    refresh = True
                except RepositoryCloneError as e:
                    _log.warning("Could not download %s: %s", repo.id, e.value)
            os.chdir(wd)
            if refresh:
                _log.debug("Refreshing package cache for %s", repo_path)
                Package.refresh_package_cache(repo_path)
    # Step 4: find packages and link to repositories
            _log.info("Looking for missing packages in local repositories.")
            for _, repo in self.repositories.iteritems():
                for id in repo.declared_packages:
                    if id in self.packages:
                        _log.debug("Binding %s to %s", id, _)
                        repo.packages.append(self.packages[id])
                        self.packages[id].repository = repo
                    elif id in missing:
                        pkg = Package.locate_offline(id, repo_path)
                        if pkg:
                            _log.debug("Found %s in clones.", id)
                            SourceFile.populate_package(pkg, self.files)
                            self.packages[id] = pkg
                            missing.remove(id)
                            repo.packages.append(pkg)
                            pkg.repository = repo
                        else:
                            _log.debug("%s was not found in clones.", id)
    # Step 5: sort packages in topological order
        self._topological_sort()
        for id in missing:
            _log.warning("Could not find package " + id)
    # Step 6: check if operating in launch mode
        self._search_launch_files()


    def _topological_sort(self):
        self._topological_packages = []
        dependencies = {}
        tier = 0
        pending = []
        emitted = []
        for name, pkg in self.packages.iteritems():
            pkg._tier = -1
            pending.append(pkg)
            dependencies[name] = set(p for p in pkg.dependencies \
                                       if p in self.packages)
        while pending:
            next_pending = []
            next_emitted = []
            for pkg in pending:
                deps = dependencies[pkg.id]
                deps.difference_update(emitted)
                if deps:
                    next_pending.append(pkg)
                else:
                    pkg._tier = tier
                    self._topological_packages.append(pkg)
                    next_emitted.append(pkg.id)
            if not next_emitted:
                # cyclic dependencies detected
                _log.warning("Cyclic dependencies detected %s", next_pending)
                for pkg in next_pending:
                    pkg._tier = tier
                    self._topological_packages.append(pkg)
                next_pending = None
            pending = next_pending
            emitted = next_emitted
            tier += 1
        self._topological_packages.sort(key = attrgetter("_tier", "id"))


    def _read_launch_listing(self, launch_files):
        _log.debug("DataManager._read_launch_listing(%s)", launch_files)
        if not launch_files:
            _log.debug("No launch files were provided.")
            return []
        if isinstance(launch_files, basestring):
            _log.debug("A single launch file was provided.")
            launch_files = (launch_files,)
        pkg_list = []
        for path in launch_files:
            path = path.split(os.sep, 1)
            if len(path) < 2:
                _log.warning("Launch files must be in the form <pkg>/path/to/file.")
                continue
            self.launch_files.append((path[0], path[1]))
            pkg_list.append(path[0])
        return pkg_list


    def _search_launch_files(self):
        _log.debug("DataManager._search_launch_files()")
        launch_files = []
        for pkg, path in self.launch_files:
            if not pkg in self.packages:
                _log.debug("Skipping %s because %s could not be found.",
                            path, pkg)
                continue
            pkg = self.packages[pkg]
            path = os.path.join(pkg.path, path)
            for sf in pkg.source_files:
                if sf.get_path() == path:
                    launch_files.append(sf)
        self.launch_files = launch_files


    def save_state(self, file_path):
        _log.debug("DataManager.save_state(%s)", file_path)
        with open(file_path, "w") as handle:
            cPickle.dump(self, handle, cPickle.HIGHEST_PROTOCOL)

    @staticmethod
    def load_state(file_path):
        _log.debug("DataManager.load_state(%s)", file_path)
        with open(file_path, "r") as handle:
            return cPickle.load(handle)


if __name__ == "__main__":
    data = DataManager()
# ----- test data -------------------------------------------------------------
    pkg = Package("A")
    pkg.dependencies.add("X")
    data.packages[pkg.id] = pkg
    pkg = Package("B")
    pkg.dependencies.add("A")
    data.packages[pkg.id] = pkg
    pkg = Package("C")
    pkg.dependencies.add("X")
    pkg.dependencies.add("A")
    data.packages[pkg.id] = pkg
    pkg = Package("D")
    pkg.dependencies.add("A")
    pkg.dependencies.add("C")
    data.packages[pkg.id] = pkg
    pkg = Package("E")
    data.packages[pkg.id] = pkg
    pkg = Package("F")
    pkg.dependencies.add("E")
    data.packages[pkg.id] = pkg
# ----- topological sort test -------------------------------------------------
    data._topological_sort()
    print data._topological_packages




#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

import json
import logging
import os

_log = logging.getLogger(__name__)


################################################################################
# Public Functions
################################################################################

def export_packages(datadir, packages):
    _log.info("Exporting package data.")
    out = os.path.join(datadir, "packages.json")
    s = "[" + ", ".join([_pkg_json(p) for _, p in packages.iteritems()]) + "]"
    with open(out, "w") as f:
        _log.debug("Writing to %s", out)
        f.write(s)

def export_rules(datadir, rules):
    _log.info("Exporting analysis rules.")
    out = os.path.join(datadir, "rules.json")
    with open(out, "w") as f:
        _log.debug("Writing to %s", out)
        json.dump([rule.__dict__ for _, rule in rules.iteritems()], f)

def export_metrics(datadir, metrics):
    _log.info("Exporting analysis metrics.")
    out = os.path.join(datadir, "metrics.json")
    with open(out, "w") as f:
        _log.debug("Writing to %s", out)
        json.dump([metric.__dict__ for _, metric in metrics.iteritems()], f)

def export_violations(datadir, packages):
    _log.info("Exporting reported rule violations.")
    for id, pkg in packages.iteritems():
        out = os.path.join(datadir, id + ".json")
        data = [_violation_json(d) for d in pkg._violations]
        for f in pkg.source_files:
            data.extend([_violation_json(d) for d in f._violations])
        with open(out, "w") as f:
            _log.debug("Writing to %s", out)
            f.write("[" + ", ".join(data) + "]")

def export_measurements(datadir, packages):
    _log.info("Exporting metrics measurements.")
    for id, pkg in packages.iteritems():
        out = os.path.join(datadir, id + ".json")
        data = [_metric_json(d) for d in pkg._metrics]
        for f in pkg.source_files:
            data.extend([_metric_json(d) for d in f._metrics])
        with open(out, "w") as f:
            _log.debug("Writing to %s", out)
            f.write("[" + ", ".join(data) + "]")

def export_configurations(datadir, packages):
    _log.info("Exporting launch configurations.")
    for id, pkg in packages.iteritems():
        out = os.path.join(datadir, id + ".json")
        data = [_config_json(d) for d in pkg._configs]
        with open(out, "w") as f:
            _log.debug("Writing to %s", out)
            f.write("[" + ", ".join(data) + "]")

def export_summary(datadir, data):
    _log.info("Exporting analysis summary.")
    out = os.path.join(datadir, "summary.json")
    with open(out, "w") as f:
        _log.debug("Writing to %s", out)
        json.dump({
            "source":           _summary_source(data.packages, data.files),
            "issues":           _summary_issues(data.repositories,
                                                data.packages, data.files),
            "components":       _summary_components(data.files, data.packages),
            "communications":   _summary_communications(data.packages)
        }, f)
    


################################################################################
# Helper Functions
################################################################################

def _summary_source(packages, files):
    langs = {}
    source_size = 0
    scripts = 0
    for _, f in files.iteritems():
        langs[f.language] = langs.get(f.language, 0) + f.size
        source_size += f.size
        if f.path.startswith("scripts" + os.path.sep):
            scripts += 1
    source_size = float(source_size)
    for lang, value in langs.iteritems():
        langs[lang] = value / source_size
    return {
        "packages": len(packages),
        "files":    len(files),
        "scripts":  scripts,
        "languages": langs
    }

def _summary_issues(repositories, packages, files):
    issues = 0
    coding = 0
    metrics = 0
    other = 0
    lines = 0
    for _, r in repositories.iteritems():
        for v in r._violations:
            any = False
            issues += 1
            if "code-standards" in v.rule.tags:
                any = True
                coding += 1
            if "metrics" in v.rule.tags:
                any = True
                metrics += 1
            if not any:
                other += 1
    for _, p in packages.iteritems():
        for v in p._violations:
            any = False
            issues += 1
            if "code-standards" in v.rule.tags:
                any = True
                coding += 1
            if "metrics" in v.rule.tags:
                any = True
                metrics += 1
            if not any:
                other += 1
    for _, f in files.iteritems():
        lines += f.lines
        for v in f._violations:
            any = False
            issues += 1
            if "code-standards" in v.rule.tags:
                any = True
                coding += 1
            if "metrics" in v.rule.tags:
                any = True
                metrics += 1
            if not any:
                other += 1
    return {
        "total":    issues,
        "coding":   coding,
        "metrics":  metrics,
        "other":    other,
        "ratio":    "{0:.2f}".format(float(issues) / lines)
    }

def _summary_components(files, packages):
    nodes = 0
    nodelets = 0
    configs = 0
    for p in packages.itervalues():
        for c in p._configs:
            configs += 1
            for n in c.nodes():
                if n.nodelet:
                    nodelets += 1
                else:
                    nodes += 1
    return {
        "launchFiles":      len([f for _, f in files.iteritems()
                                 if f.language == "launch"]),
        "nodes":            nodes,
        "nodelets":         nodelets,
        "parameterFiles":   None,
        "configurations":   configs
    }

def _summary_communications(packages):
    topics = 0
    remappings = 0
    for p in packages.itervalues():
        for c in p._configs:
            remaps = dict(c.resources.remaps)
            for n in c.nodes():
                remaps.update(n.remaps)
            remappings += len(remaps)
            topics += len(c.resources.get_topics())
            topics += len(c.resources.get_services())
    return {
        "topics":       topics,
        "remappings":   remappings,
        "messages":     None,
        "services":     None,
        "actions":      None
    }

def _pkg_json(pkg):
    s = '{"id": "' + pkg.id + '", '
    s += '"metapackage": ' + json.dumps(pkg.isMetapackage)
    s += ', "description": "' + _escaped(pkg.description)
    s += '", "wiki": ' + json.dumps(pkg.website)
    s += ', "repository": ' + json.dumps(pkg.vcs_url)
    s += ', "bugTracker": ' + json.dumps(pkg.bug_url)
    s += ', "authors": ' + json.dumps([p.name for p in pkg.authors])
    s += ', "maintainers": ' + json.dumps([p.name for p in pkg.maintainers])
    s += ', "dependencies": ' + json.dumps(list(pkg.dependencies))
    s += ', "size": ' + "{0:.2f}".format(pkg.size / 1000.0)
    s += ', "lines": ' + str(pkg.lines)
    analysis = {}
    violations = {}
    metrics = {}
    for datum in pkg._violations:
        violations[datum.rule.id] = violations.get(datum.rule.id, 0) + 1
    for f in pkg.source_files:
        for datum in f._violations:
            violations[datum.rule.id] = violations.get(datum.rule.id, 0) + 1
    for datum in pkg._metrics:
        metrics[datum.metric.id] = datum.value
    analysis["violations"] = violations
    analysis["metrics"] = metrics
    s += ', "analysis": ' + json.dumps(analysis) + "}"
    return s


def _violation_json(datum):
    s = '{"rule": "' + datum.rule.id + '", '
    if datum.rule.scope == "file" \
            or datum.rule.scope == "function" \
            or datum.rule.scope == "class":
        s += '"file": "' + datum.scope.name + '", '
        try:
            s += '"line": ' + json.dumps(datum.line) + ", "
            s += '"function": ' + json.dumps(datum.function) + ", "
            s += '"class": ' + json.dumps(datum.class_) + ", "
        except AttributeError as e:
            _log.debug("_violation_json %s", e)
    s += '"comment": "' + _escaped(datum.details or "") + '"}'
    return s

def _metric_json(datum):
    s = '{"metric": "' + datum.metric.id + '", '
    if datum.metric.scope == "file" \
            or datum.metric.scope == "function" \
            or datum.metric.scope == "class":
        s += '"file": "' + datum.scope.name + '", '
        try:
            s += '"line": ' + json.dumps(datum.line) + ", "
            s += '"function": ' + json.dumps(datum.function) + ", "
            s += '"class": ' + json.dumps(datum.class_) + ", "
        except AttributeError as e:
            _log.debug("_metric_json %s", e)
    s += '"value": ' + str(datum.value) + "}"
    return s

def _config_json(config):
    i = config.name + "-" + str(hash(config.package + config.name) % 10000000)
    r = dict(config.resources.remaps)
    for n in config.nodes():
        r.update(n.remaps)
    s = ('{"id": "' + i + '", "name": "' + config.name + '", "collisions": '
         + str(config.resources.n_collisions) + ', "remaps": '
         + str(len(r)) + ', "dependencies": '
         + json.dumps(list(config.pkg_depends)) + ', "environment": '
         + json.dumps(list(config.env_depends)) + ', "nodes": '
         + json.dumps(map(_node_json, config.nodes())) + "}")
    return s

def _node_json(node):
    name = lambda t: t[0].full_name
    return {
        "name":         node.full_name,
        "type":         node.reference,
        "args":         node.argv,
        "publishers":   map(name, node.publishers),
        "subscribers":  map(name, node.subscribers),
        "servers":      map(name, node.servers),
        "clients":      map(name, node.clients)
    }


def _escaped(s):
    return s.replace('"', '\\"').replace("\n", " ").replace("<", "&lt;")\
            .replace(">", "&gt;").replace("&", "&amp;").replace("\t", " ")\
            .replace("\\012", " ")




#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

import importlib
import logging
import os
import sys
import yaml

_log = logging.getLogger(__name__)

class MalformedManifestError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class AnalysisInterface(object):
    def __init__(self, module, languages):
        self.module     = module
        self.languages  = set(languages)
        self.state      = None
        self.f_analysis = hasattr(module, "file_analysis")
        self.p_analysis = hasattr(module, "package_analysis")
        self.r_analysis = hasattr(module, "repository_analysis")

    def analyse_file(self, iface, scope):
        _log.debug("Plugin.analyse_file: " + scope.id)
        if self.f_analysis and scope.language in self.languages:
            _log.debug("Calling module.file_analysis")
            self.module.file_analysis(iface, scope)

    def analyse_package(self, iface, scope):
        _log.debug("Plugin.analyse_package: " + scope.id)
        if self.p_analysis:
            _log.debug("Calling module.package_analysis")
            self.module.package_analysis(iface, scope)

    def analyse_repository(self, iface, scope):
        # Note: path may be None if the repository was not downloaded
        # TODO: what action to take on None path
        _log.debug("Plugin.analyse_repository: " + scope.id)
        if self.r_analysis:
            _log.debug("Calling module.repository_analysis")
            self.module.repository_analysis(iface, scope)

    def pre_analysis(self):
        _log.debug("Plugin.pre_analysis")
        try:
            _log.debug("Calling module.pre_analysis")
            self.state = self.module.pre_analysis()
        except AttributeError as e:
            _log.debug("Module does not perform initialisation.")

    def post_analysis(self, iface):
        _log.debug("Plugin.post_analysis")
        try:
            _log.debug("Calling module.post_analysis")
            self.module.post_analysis(iface)
        except AttributeError as e:
            _log.debug("Module does not perform finalisation.")


class ProcessingInterface(object):
    def __init__(self, module):
        self.module         = module
        self.state          = None
        self.f_violations   = hasattr(module, "process_file_violation")
        self.f_metrics      = hasattr(module, "process_file_metric")
        self.p_violations   = hasattr(module, "process_package_violation")
        self.p_metrics      = hasattr(module, "process_package_metric")
        self.r_violations   = hasattr(module, "process_repository_violation")
        self.r_metrics      = hasattr(module, "process_repository_metric")

    def process_file(self, iface, scope, violations, metrics):
        # Receives a copy of the issues because the actual lists may change
        _log.debug("Plugin.process_file: " + scope.id)
        if self.f_violations:
            _log.debug("Calling module.process_file_violation")
            for datum in violations:
                self.module.process_file_violation(iface, datum)
        if self.f_metrics:
            _log.debug("Calling module.process_file_metric")
            for datum in metrics:
                self.module.process_file_metric(iface, datum)

    def process_package(self, iface, scope, violations, metrics):
        # Receives a copy of the issues because the actual lists may change
        _log.debug("Plugin.process_package: " + scope.id)
        if self.p_violations:
            _log.debug("Calling module.process_package_violation")
            for datum in violations:
                self.module.process_package_violation(iface, datum)
        if self.p_metrics:
            _log.debug("Calling module.process_package_metric")
            for datum in metrics:
                self.module.process_package_metric(iface, datum)

    def process_repository(self, iface, scope, violations, metrics):
        # Receives a copy of the issues because the actual lists may change
        _log.debug("Plugin.process_repository: " + scope.id)
        if self.r_violations:
            _log.debug("Calling module.process_repository_violation")
            for datum in violations:
                self.module.process_repository_violation(iface, datum)
        if self.r_metrics:
            _log.debug("Calling module.process_repository_metric")
            for datum in metrics:
                self.module.process_repository_metric(iface, datum)

    def pre_process(self):
        _log.debug("Plugin.pre_process")
        try:
            _log.debug("Calling module.pre_process")
            self.state = self.module.pre_process()
        except AttributeError as e:
            _log.debug("Module does not perform initialisation.")

    def post_process(self, iface):
        _log.debug("Plugin.post_process")
        try:
            _log.debug("Calling module.post_process")
            self.module.post_process(iface)
        except AttributeError as e:
            _log.debug("Module does not perform finalisation.")


class ExportInterface(object):
    pass


class Plugin(object):
    def __init__(self, name, dir):
        self.name       = name
        self.version    = "0.1"
        self.path       = dir
        self.rules      = None
        self.metrics    = None
        self.analysis   = None
        self.process    = None
        self.export     = None
        self.tmp_path   = None

    def load(self, common_rules = None, common_metrics = None):
        _log.debug("Plugin.load")
        manifest = os.path.join(self.path, "plugin.yaml")
        _log.debug("Plugin manifest at " + manifest)
        with open(manifest, "r") as openfile:
            manifest = yaml.load(openfile)
        if not "version" in manifest or not "name" in manifest or \
                manifest["name"] != self.name:
            raise MalformedManifestError("Malformed plugin manifest: " + \
                    self.name)
        self.version    = manifest["version"]
        self.rules      = manifest.get("rules", {})
        self.metrics    = manifest.get("metrics", {})
        _log.debug("Loaded %s [%s]", self.name, self.version)
        if common_rules:
            rm = [id for id in self.rules if id in common_rules]
            for id in rm:
                _log.warning("Plugin %s cannot override %s", self.name, id)
                del self.rules[id]
        if common_metrics:
            rm = [id for id in self.metrics if id in common_metrics]
            for id in rm:
                _log.warning("Plugin %s cannot override %s", self.name, id)
                del self.metrics[id]
        _log.info("Loading plugin script.")
        _log.debug("Plugin script at %s", self.path)
        module = importlib.import_module(self.name + ".plugin",
                                         package = self.name)
        languages = manifest.get("languages", [])
        self.analysis = AnalysisInterface(module, languages)
        self.process = ProcessingInterface(module)
        self.export = ExportInterface()



def load_plugins(root, whitelist = None, blacklist = None):
    _log.debug("load_plugins(%s, %s, %s)", root, whitelist, blacklist)
    plugins = []
    filter = []
    mode = 0
    if whitelist:
        filter = whitelist
        mode = 1
    elif blacklist:
        filter = blacklist
        mode = -1
    sys.path.insert(0, root)
    for item in os.listdir(root):
        if (mode > 0 and not item in filter) or (mode < 0 and item in filter) \
                or item.startswith(".") or item == "haros_util":
            continue
        d = os.path.join(root, item)
        _log.debug("Checking for plugins at %s", d)
        if os.path.isdir(d) \
                and os.path.isfile(os.path.join(d, "plugin.yaml")) \
                and os.path.isfile(os.path.join(d, "plugin.py")):
            plugin = Plugin(item, d)
            try:
                plugin.load()
                plugins.append(plugin)
            except MalformedManifestError as e:
                _log.warning(e.value)
            except ImportError as e:
                _log.error("Failed to import %s; %s", item, e)
    sys.path.pop(0)
    return plugins




#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

import logging
import os
import subprocess
import threading

from distutils.dir_util import copy_tree
from pkg_resources import Requirement, resource_filename
from shutil import rmtree

try: 
    # Python 3
    from http.server import HTTPServer, BaseHTTPRequestHandler
except ImportError: 
    # Python 2
    import SimpleHTTPServer
    from BaseHTTPServer import HTTPServer
    from SimpleHTTPServer import SimpleHTTPRequestHandler

    class BaseHTTPRequestHandler(SimpleHTTPRequestHandler):
        def end_headers(self):
            self.send_my_headers()
            SimpleHTTPRequestHandler.end_headers(self)

        def send_my_headers(self):
            self.send_header("Cache-Control", "no-cache, no-store, must-revalidate")
            self.send_header("Pragma", "no-cache")
            self.send_header("Expires", "0")


_log = logging.getLogger(__name__)


def install(dst):
    if os.path.exists(dst):
        rmtree(dst)
    if not os.path.exists(dst):
        data_dir = os.path.join(dst, "data")
        _log.info("Creating %s", dst)
        os.mkdir(dst)
        _log.info("Copying viz files.")
        src = resource_filename(Requirement.parse("haros"), "harosviz")
        copy_tree(src, dst)
        _log.info("Creating %s", data_dir)
        os.mkdir(data_dir)
        _log.info("Creating %s", os.path.join(data_dir, "compliance"))
        os.mkdir(os.path.join(data_dir, "compliance"))
        _log.info("Creating %s", os.path.join(data_dir, "metrics"))
        os.mkdir(os.path.join(data_dir, "metrics"))
        _log.info("Creating %s", os.path.join(data_dir, "models"))
        os.mkdir(os.path.join(data_dir, "models"))


def serve(directory, host_str):
    host = host_str.split(":")
    if len(host) != 2:
        raise RuntimeError("Invalid host:port provided: " + host_str)
    wd = os.getcwd()
    try:
        os.chdir(directory)
        server = HTTPServer((host[0], int(host[1])), BaseHTTPRequestHandler)
        print "[HAROS] Serving visualisation at", host_str
        thread = threading.Thread(target = server.serve_forever)
        thread.deamon = True
        thread.start()
        _log.info("Starting web browser process.")
        p = subprocess.Popen(["python", "-m", "webbrowser",
                        "-t", "http://" + host_str])
        raw_input("[HAROS] Press enter to shutdown the viz server:")
        return True
    except ValueError as e:
        _log.error("Invalid port for the viz server %s", host[1])
        return False
    finally:
        server.shutdown()
        os.chdir(wd)
        _log.debug("Killing web browser process.")
        p.kill()




#Copyright (c) 2016 Andre Santos
#
#Permission is hereby granted, free of charge, to any person obtaining a copy
#of this software and associated documentation files (the "Software"), to deal
#in the Software without restriction, including without limitation the rights
#to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#copies of the Software, and to permit persons to whom the Software is
#furnished to do so, subject to the following conditions:

#The above copyright notice and this permission notice shall be included in
#all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#THE SOFTWARE.

"""
HAROS directory (data folder) structure:

+ ~/.haros
|-- index.yaml
|-+ plugins
  |-+ ...
|-+ repositories
  |-+ ...
|-+ viz
  |-+ ...
  |-+ data
    |-- packages.json
    |-- rules.json
    |-- summary.json
    |-+ compliance
      |-- ...
    |-+ metrics
      |-- ...
  |-- index.html
|-+ export
  |-- metrics.csv
"""

# start with init
# init creates the default data dir
# viz is copied to init dir

# analysis grabs previous db and index from data dir
# analysis may accept import option to use another db

# export receives a dir where it will generate the export files
# export still generates files in the data dir
# export uses db from analysis step or loads it from data dir

# viz uses data from the data dir


import argparse
import logging
import os
import subprocess
import sys

from shutil import copyfile
from pkg_resources import Requirement, resource_filename

from .data_manager import DataManager
from . import plugin_manager as plugman
from . import analysis_manager as anaman
from . import export_manager as expoman
from . import visualiser as viz


HAROS_DIR       = os.path.join(os.path.expanduser("~"), ".haros")
REPOSITORY_DIR  = os.path.join(HAROS_DIR, "repositories")
EXPORT_DIR      = os.path.join(HAROS_DIR, "export")
PLUGIN_DIR      = os.path.join(HAROS_DIR, "plugins")
VIZ_DIR         = os.path.join(HAROS_DIR, "viz")
VIZ_DATA_DIR    = os.path.join(VIZ_DIR, "data")
DB_PATH         = os.path.join(HAROS_DIR, "haros.db")
LOG_PATH        = os.path.join(HAROS_DIR, "log.txt")
PLUGIN_REPOSITORY = "https://github.com/git-afsantos/haros_plugins.git"

_log = logging.getLogger()

# Options:
#   --debug sets the logging level to debug
#   haros init
#       initialises the data directory
#   haros analyse [args]
#       runs update, analysis and export
#       -k  keep previous analyses (run analysis only for new packages)
#       -r  also register and analyse repositories
#       -w  whitelist plugins
#       -b  blacklist plugins
#       -p  package filter
#       -d  db to import
#   haros export [args]
#       runs export only
#       -d export dir (output)
#   haros viz [args]
#       runs visualiser only
#       -s host
#       -d export dir (input)
#   haros full [args]
#       full run (analyse + viz)
def parse_arguments(argv):
    parser = argparse.ArgumentParser(prog="haros",
            description="ROS quality assurance.")
    parser.add_argument("--debug", action = "store_true",
                        help = "set debug logging")
    subparsers = parser.add_subparsers()

    parser_init = subparsers.add_parser("init")
    parser_init.set_defaults(func = command_init)

    parser_full = subparsers.add_parser("full")
    parser_full.add_argument("-r", "--repositories", dest = "use_repos",
                             action = "store_true", help = "use repositories")
    parser_full.add_argument("-s", "--server-host", dest = "host",
                             default = "localhost:8080",
                             help = "visualisation host " \
                                    "(default: \"localhost:8080\")")
    parser_full.add_argument("-p", "--package-index", dest = "pkg_filter",
                             help = "package index file")
    group = parser_full.add_mutually_exclusive_group()
    group.add_argument("-w", "--whitelist", nargs = "*", dest = "whitelist",
                       help = "execute only these plugins")
    group.add_argument("-b", "--blacklist", nargs = "*", dest = "blacklist",
                       help = "skip these plugins")
    parser_full.set_defaults(func = command_full)

    parser_analyse = subparsers.add_parser("analyse")
    parser_analyse.add_argument("-r", "--repositories", dest = "use_repos",
                                action = "store_true",
                                help = "use repositories")
    parser_analyse.add_argument("-p", "--package-index", dest = "pkg_filter",
                                help = "package index file")
    group = parser_analyse.add_mutually_exclusive_group()
    group.add_argument("-w", "--whitelist", nargs = "*", dest = "whitelist",
                       help="execute only these plugins")
    group.add_argument("-b", "--blacklist", nargs = "*", dest = "blacklist",
                       help="skip these plugins")
    parser_analyse.set_defaults(func = command_analyse)

    parser_export = subparsers.add_parser("export")
    parser_export.add_argument("data_dir", metavar = "dir",
                               help = "where to export data")
    parser_export.set_defaults(func = command_export)

    parser_viz = subparsers.add_parser("viz")
    parser_viz.add_argument("-s", "--server-host", dest = "host",
                            default = "localhost:8080",
                            help = "visualisation host " \
                                   "(default: \"localhost:8080\")")
    parser_viz.set_defaults(func = command_viz)

    return parser.parse_args() if argv is None else parser.parse_args(argv)


def _check_haros_directory():
    if not os.path.isdir(HAROS_DIR):
        raise RuntimeError("HAROS directory was not initialised.")

def _empty_dir(dir_path):
    _log.debug("Emptying directory %s", dir_path)
    for f in os.listdir(dir_path):
        path = os.path.join(dir_path, f)
        if os.path.isfile(path):
            _log.debug("Removing file %s", path)
            os.unlink(path)

def command_init(args):
    print "[HAROS] Creating directories..."
    if os.path.exists(HAROS_DIR) and not os.path.isdir(HAROS_DIR):
        raise RuntimeError("Could not init; " + HAROS_DIR \
                           + " already exists and is not a directory.")
    if not os.path.exists(HAROS_DIR):
        _log.info("Creating %s", HAROS_DIR)
        os.makedirs(HAROS_DIR)
        _log.info("Creating %s", os.path.join(HAROS_DIR, "index.yaml"))
        with open(os.path.join(HAROS_DIR, "index.yaml"), "w") as f:
            f.write("%YAML 1.1\n---\npackages: []\n")
    if not os.path.exists(REPOSITORY_DIR):
        _log.info("Creating %s", REPOSITORY_DIR)
        os.mkdir(REPOSITORY_DIR)
    if not os.path.exists(EXPORT_DIR):
        _log.info("Creating %s", EXPORT_DIR)
        os.mkdir(EXPORT_DIR)
    viz.install(VIZ_DIR)
    if not os.path.exists(PLUGIN_DIR):
        _log.info("Creating %s", PLUGIN_DIR)
        os.mkdir(PLUGIN_DIR)
        _log.info("Cloning plugin repository.")
        subprocess.check_call(["git", "clone", PLUGIN_REPOSITORY, PLUGIN_DIR])
    else:
        _log.info("Updating plugin repository.")
        wd = os.getcwd()
        os.chdir(PLUGIN_DIR)
        subprocess.check_call(["git", "pull"])
        os.chdir(wd)


def command_full(args):
    if command_analyse(args):
        return command_viz(args)
    return False


def command_analyse(args):
    _check_haros_directory()
    _log.debug("Creating new data manager.")
    dataman = DataManager()
    # path = os.path.join(args.datadir, "haros.db")
    # if os.path.isfile(path):
        # dataman = DataManager.load_state()
    print "[HAROS] Indexing source code..."
    path = args.pkg_filter \
           if args.pkg_filter and os.path.isfile(args.pkg_filter) \
           else os.path.join(HAROS_DIR, "index.yaml")
    _log.debug("Package index file %s", path)
    if not os.path.isfile(path):
        raise RuntimeError("There is no package index file. Aborting.")
    dataman.index_source(path, REPOSITORY_DIR, args.use_repos)
    if not dataman.packages:
        _log.warning("There are no packages to analyse.")
        return False
    print "[HAROS] Loading common definitions..."
    path = resource_filename(Requirement.parse("haros"),
                             "haros/definitions.yaml")
    dataman.load_definitions(path)
    print "[HAROS] Loading plugins..."
    plugins = plugman.load_plugins(PLUGIN_DIR, args.whitelist, args.blacklist)
    if not plugins:
        _log.warning("There are no analysis plugins.")
        return False
    for plugin in plugins:
        dataman.extend_definitions(plugin.name, plugin.rules, plugin.metrics)
    print "[HAROS] Running analysis..."
    _empty_dir(EXPORT_DIR)
    anaman.run_analysis_and_processing(HAROS_DIR, plugins, dataman, EXPORT_DIR)
    print "[HAROS] Saving analysis results..."
    dataman.save_state(DB_PATH)
    command_export(args, dataman)
    return True


def command_export(args, dataman = None):
    _check_haros_directory()
    print "[HAROS] Exporting analysis results..."
    if dataman:
        _log.debug("Exporting on-memory data manager.")
        json_path   = VIZ_DATA_DIR
        csv_path    = EXPORT_DIR
        db_path     = None
        _empty_dir(os.path.join(VIZ_DATA_DIR, "compliance"))
        _empty_dir(os.path.join(VIZ_DATA_DIR, "metrics"))
    elif os.path.isdir(args.data_dir):
        _log.debug("Exporting data manager from file.")
        if os.path.isfile(DB_PATH):
            dataman = DataManager.load_state(DB_PATH)
        else:
            _log.warning("There is no analysis data to export.")
            return False
        json_path = os.path.join(args.data_dir, "json")
        if not os.path.exists(json_path):
            _log.info("Creating directory %s", json_path)
            os.mkdir(json_path)
        csv_path = os.path.join(args.data_dir, "csv")
        if not os.path.exists(csv_path):
            _log.info("Creating directory %s", csv_path)
            os.mkdir(csv_path)
        db_path = os.path.join(args.data_dir, "haros.db")
    else:
        _log.error("%s is not a directory!", args.data_dir)
        return False
    expoman.export_packages(json_path, dataman.packages)
    expoman.export_rules(json_path, dataman.rules)
    expoman.export_metrics(json_path, dataman.metrics)
    expoman.export_summary(json_path, dataman)
    path = os.path.join(json_path, "compliance")
    if not os.path.exists(path):
        _log.info("Creating directory %s", path)
        os.mkdir(path)
    expoman.export_violations(path, dataman.packages)
    path = os.path.join(json_path, "metrics")
    if not os.path.exists(path):
        _log.info("Creating directory %s", path)
        os.mkdir(path)
    expoman.export_measurements(path, dataman.packages)
    path = os.path.join(json_path, "models")
    if not os.path.exists(path):
        _log.info("Creating directory %s", path)
        os.mkdir(path)
    expoman.export_configurations(path, dataman.packages)
    if db_path:
        _log.debug("Copying DB from %s to %s", DB_PATH, db_path)
        copyfile(DB_PATH, db_path)
    return True


def command_viz(args):
    _check_haros_directory()
    viz.serve(VIZ_DIR, args.host)


def main(argv = None):
    args = parse_arguments(argv)
    if args.debug:
        logging.basicConfig(filename = LOG_PATH, filemode = "w",
                            level = logging.DEBUG)
    else:
        logging.basicConfig(level = logging.WARNING)
    try:
        _log.info("Executing selected command.")
        args.func(args)
        return 0
    except RuntimeError as err:
        _log.error(str(err))
        return 1
